@@ -125,6 +125,7 @@ class History:
         self.last_cmd_out = None
         self.hist_size = None
         self.hist_units = None
+        self.remember_history = True
 
     def __len__(self):
         """Return the number of items in current session."""
@@ -208,6 +209,12 @@ class History:
         """
         pass
 
+    def clear(self):
+        """Clears the history of the current session from both the disk and
+        memory.
+        """
+        pass
+
 #
 # dummy
 #
@@ -241,16 +248,20 @@ class DummyHistory(History):
 os = _LazyModule.load('os', 'os')
 sys = _LazyModule.load('sys', 'sys')
 time = _LazyModule.load('time', 'time')
-json = _LazyModule.load('json', 'json')
 builtins = _LazyModule.load('builtins', 'builtins')
 # amalgamated collections
 threading = _LazyModule.load('threading', 'threading')
 cabc = _LazyModule.load('collections', 'collections.abc', 'cabc')
+try:
+    import ujson as json
+except ImportError:
+    import json  # type: ignore
+
 # amalgamated xonsh.history.base
 xt = _LazyModule.load('xonsh', 'xonsh.tools', 'xt')
 xlj = _LazyModule.load('xonsh', 'xonsh.lazyjson', 'xlj')
 uptime = _LazyModule.load('xonsh', 'xonsh.xoreutils.uptime', 'uptime')
-def _xhj_gc_commands_to_rmfiles(hsize, files) -> ([], int):
+def _xhj_gc_commands_to_rmfiles(hsize, files):
     """Return number of units and list of history files to remove to get under the limit,
 
     Parameters:
@@ -397,7 +408,7 @@ class JsonHistoryGC(threading.Thread):
         else:
             print(
                 f"Warning: History garbage collection would discard more history ({size_over} {units}) than it would keep ({hsize}).\n"
-                "Not removing any history for now. Either increase your limit ($XONSH_HIST_SIZE), or run `history gc --force`."
+                "Not removing any history for now. Either increase your limit ($XONSH_HISTORY_SIZE), or run `history gc --force`."
             )
 
     def files(self, only_unlocked=False):
@@ -415,6 +426,7 @@ class JsonHistoryGC(threading.Thread):
         boot = uptime.boottime()
         fs = _xhj_get_history_files(sort=False)
         files = []
+        time_start = time.time()
         for f in fs:
             try:
                 cur_file_size = os.path.getsize(f)
@@ -437,11 +449,16 @@ class JsonHistoryGC(threading.Thread):
                 # info: file size, closing timestamp, number of commands, filename
                 ts = lj.get("ts", (0.0, None))
                 files.append(
-                    (ts[1] or ts[0], len(lj.sizes["cmds"]) - 1, f, cur_file_size,),
+                    (ts[1] or ts[0], len(lj.sizes["cmds"]) - 1, f, cur_file_size)
                 )
                 lj.close()
                 if xonsh_debug:
-                    print(f"... Enumerated {len(files):7d} history files.\r", end="")
+                    time_lag = time.time() - time_start
+                    print(
+                        f"[history.{json.__name__}] Enumerated {len(files):,d} history files for {time_lag:0.4f}s.\r",
+                        end="",
+                        file=sys.stderr,
+                    )
             except (IOError, OSError, ValueError):
                 continue
         files.sort()  # this sorts by elements of the tuple,
@@ -497,6 +514,7 @@ class JsonHistoryFlusher(threading.Thread):
                 if self.skip is not None:
                     self.skip(1)
                 continue
+
             cmds.append(cmd)
             last_inp = cmd["inp"]
         with open(self.filename, "r", newline="\n") as f:
@@ -538,6 +556,9 @@ class JsonCommandField(cabc.Sequence):
         return len(self.hist)
 
     def __getitem__(self, key):
+        if not self.hist.remember_history:
+            return ""
+
         size = len(self)
         if isinstance(key, slice):
             return [self[i] for i in range(*key.indices(size))]
@@ -636,14 +657,30 @@ class JsonHistory(History):
             added to the history list. It should contain the keys ``inp``,
             ``rtn`` and ``ts``. These key names mirror the same names defined
             as instance variables in the ``HistoryEntry`` class.
+            Additionally, an optional key ``spc`` may be present which will
+            affect commands from being stored if ignorespace is in $HISTCONTROL.
 
         Returns
         -------
         hf : JsonHistoryFlusher or None
             The thread that was spawned to flush history
         """
+        if not self.remember_history:
+            return
+
+        opts = builtins.__xonsh__.env.get("HISTCONTROL")
+        skipped_by_ignore_space = "ignorespace" in opts and cmd.get("spc")
+        if skipped_by_ignore_space:
+            return None
+
         self.buffer.append(cmd)
         self._len += 1  # must come before flushing
+
+        try:
+            del cmd["spc"]
+        except KeyError:
+            pass
+
         if len(self.buffer) >= self.buffersize:
             hf = self.flush()
         else:
@@ -664,6 +701,7 @@ class JsonHistory(History):
         hf : JsonHistoryFlusher or None
             The thread that was spawned to flush history
         """
+        # Implicitly covers case of self.remember_history being False.
         if len(self.buffer) == 0:
             return
 
@@ -706,7 +744,7 @@ class JsonHistory(History):
                 continue
             try:
                 commands = json_file.load()["cmds"]
-            except json.decoder.JSONDecodeError:
+            except (json.decoder.JSONDecodeError, ValueError):
                 # file is corrupted somehow
                 if builtins.__xonsh__.env.get("XONSH_DEBUG") > 0:
                     msg = "xonsh history file {0!r} is not valid JSON"
@@ -738,6 +776,21 @@ class JsonHistory(History):
             while self.gc.is_alive():  # while waiting for gc.
                 time.sleep(0.1)  # don't monopolize the thread (or Python GIL?)
 
+    def clear(self):
+        """Clears the current session's history from both memory and disk."""
+
+        # Wipe history from memory. Keep sessionid and other metadata.
+        self.buffer = []
+        self.tss = JsonCommandField("ts", self)
+        self.inps = JsonCommandField("inp", self)
+        self.outs = JsonCommandField("out", self)
+        self.rtns = JsonCommandField("rtn", self)
+        self._len = 0
+        self._skipped = 0
+
+        # Flush empty history object to disk, overwriting previous data.
+        self.flush()
+
 #
 # sqlite
 #
@@ -745,7 +798,7 @@ class JsonHistory(History):
 """Implements the xonsh history backend via sqlite3."""
 # amalgamated builtins
 # amalgamated collections
-# amalgamated json
+json = _LazyModule.load('json', 'json')
 # amalgamated os
 sqlite3 = _LazyModule.load('sqlite3', 'sqlite3')
 # amalgamated sys
@@ -808,6 +861,13 @@ def _xh_sqlite_create_history_table(cursor):
         except sqlite3.OperationalError:
             pass
 
+        # add index on inp. since we query when erasedups is True
+        cursor.execute(
+            f"""\
+CREATE INDEX IF NOT EXISTS  idx_inp_history
+ON {XH_SQLITE_TABLE_NAME}(inp);"""
+        )
+
         # mark that this function ran for this session
         setattr(XH_SQLITE_CACHE, XH_SQLITE_CREATED_SQL_TBL, True)
 
@@ -981,14 +1041,12 @@ class SqliteHistory(History):
         setattr(XH_SQLITE_CACHE, XH_SQLITE_CREATED_SQL_TBL, False)
 
     def append(self, cmd):
+        if not self.remember_history:
+            return
         envs = builtins.__xonsh__.env
         inp = cmd["inp"].rstrip()
         self.inps.append(inp)
-        store_stdout = envs.get("XONSH_STORE_STDOUT", False)
-        if store_stdout:
-            self.outs.append(cmd.get("out"))
-        else:
-            self.outs.append(None)
+        self.outs.append(cmd.get("out"))
         self.rtns.append(cmd["rtn"])
         self.tss.append(cmd.get("ts", (None, None)))
 
@@ -999,11 +1057,19 @@ class SqliteHistory(History):
         if "ignoreerr" in opts and cmd["rtn"] != 0:
             # Skipping failed cmd
             return
+        if "ignorespace" in opts and cmd.get("spc"):
+            # Skipping cmd starting with space
+            return
+
+        try:
+            del cmd["spc"]
+        except KeyError:
+            pass
         self._last_hist_inp = inp
         xh_sqlite_append_history(
             cmd,
             str(self.sessionid),
-            store_stdout,
+            store_stdout=envs.get("XONSH_STORE_STDOUT", False),
             filename=self.filename,
             remove_duplicates=("erasedups" in opts),
         )
@@ -1038,6 +1104,21 @@ class SqliteHistory(History):
             while self.gc.is_alive():
                 continue
 
+    def clear(self):
+        """Clears the current session's history from both memory and disk."""
+        # Wipe memory
+        self.inps = []
+        self.rtns = []
+        self.outs = []
+        self.tss = []
+
+        # Wipe the current session's entries from the database.
+        sql = "DELETE FROM xonsh_history WHERE sessionid = ?"
+        with _xh_sqlite_get_conn(filename=self.filename) as conn:
+            c = conn.cursor()
+            _xh_sqlite_create_history_table(c)
+            c.execute(sql, (str(self.sessionid),))
+
 #
 # main
 #
@@ -1265,7 +1346,18 @@ def _XH_HISTORY_SESSIONS():
     }
 
 
-_XH_MAIN_ACTIONS = {"show", "id", "file", "info", "diff", "gc", "flush"}
+_XH_MAIN_ACTIONS = {
+    "show",
+    "id",
+    "file",
+    "info",
+    "diff",
+    "gc",
+    "flush",
+    "off",
+    "on",
+    "clear",
+}
 
 
 @functools.lru_cache()
@@ -1395,6 +1487,17 @@ def _xh_create_parser():
     # 'flush' subcommand
     subp.add_parser("flush", help="flush the current history to disk")
 
+    # 'off' subcommand
+    subp.add_parser("off", help="history will not be saved for this session")
+
+    # 'on' subcommand
+    subp.add_parser(
+        "on", help="history will be saved for the rest of the session (default)"
+    )
+
+    # 'clear' subcommand
+    subp.add_parser("clear", help="one-time wipe of session history")
+
     return p
 
 
@@ -1458,6 +1561,18 @@ def history_main(
         hf = hist.flush()
         if isinstance(hf, threading.Thread):
             hf.join()
+    elif ns.action == "off":
+        if hist.remember_history:
+            hist.clear()
+            hist.remember_history = False
+            print("History off", file=sys.stderr)
+    elif ns.action == "on":
+        if not hist.remember_history:
+            hist.remember_history = True
+            print("History on", file=sys.stderr)
+    elif ns.action == "clear":
+        hist.clear()
+        print("History cleared", file=sys.stderr)
     else:
         print("Unknown history action {}".format(ns.action), file=sys.stderr)
 
